@@ -73,11 +73,23 @@ ulog <- function(...) invisible(.Call(Rserve_ulog, paste(..., collapse="\n", sep
 ocap <- function(fun, name=deparse(substitute(fun)))
   .Call(Rserve_oc_register, fun, name)
 
-Rserve.eval <- function(what, where=.GlobalEnv, last.value=FALSE, exp.value=FALSE, context=NULL)
-    .Call(Rserve_eval, what, where, last.value, exp.value, context)
+.save.condition <- function(cond)
+  .Call(Rserve_set_last_condition, cond)
+
+Rserve.eval <- function(what, where=.GlobalEnv, last.value=FALSE, exp.value=FALSE,
+	    context=NULL, handlers=list(error=.save.condition))
+    .Call(Rserve_eval, what, where, last.value, exp.value, context, handlers)
 
 Rserve.context <- function(what)
     if (missing(what)) .Call(Rserve_get_context) else .Call(Rserve_set_context, what)
 
+.persistence <- new.env()
+
+Rserve.set.http.request <- function(what) {
+    if (!is.null(what) && !is.symbol(what))
+        .persistence$http.request <- what
+    invisible(.Call(Rserve_set_http_request_fn, what))
+}
+
 resolve.ocap <- function(ocap)
   .Call(Rserve_oc_resolve, ocap)
