@@ -56,11 +56,12 @@ struct args {
 	server_t *srv; /* server that instantiated this connection */
     SOCKET s;
 	SOCKET ss;
+	int msg_id;
 	void *res1, *res2;
 	/* the following entries are not populated by Rserve but can be used by server implemetations */
 	char *buf, *sbuf;
 	int   ver, bp, bl, sp, sl, flags;
-	long  l1, l2;
+	size_t l1, l2;
 	/* The following fields are informational, populated by Rserve */
     SAIN sa;
     int ucix;
@@ -311,6 +312,18 @@ static void fin_request(args_t *c) {
 		c->attr |= CONNECTION_CLOSE;
 }
 
+static SEXP s_http_request_fn;
+
+static void http_set_request_fn(SEXP sFn) {
+	s_http_request_fn = sFn ? sFn : install(".http.request");
+}
+
+/* API */
+SEXP Rserve_set_http_request_fn(SEXP sFn) {
+	http_set_request_fn((sFn == R_NilValue) ? 0 : sFn);
+	return s_http_request_fn;
+}
+
 /* process a request by calling the httpd() function in R */
 static void process_request(args_t *c)
 {
@@ -332,6 +345,8 @@ static void process_request(args_t *c)
 		*(s++) = 0;
 		query = s;
     }
+	if (!s_http_request_fn)
+		s_http_request_fn = install(".http.request");
     uri_decode(c->url); /* decode the path part */
     {   /* construct "try(httpd(url, query, body, headers), silent=TRUE)" */
 		SEXP sTrue = PROTECT(ScalarLogical(TRUE));
@@ -341,7 +356,7 @@ static void process_request(args_t *c)
 		SEXP sArgs = PROTECT(list4(mkString(c->url), sQuery, sBody, sReqHeaders));
 		SEXP sTry = install("try");
 		SEXP y, x = PROTECT(lang3(sTry,
-								  LCONS(install(".http.request"), sArgs),
+								  LCONS(s_http_request_fn, sArgs),
 								  sTrue));
 		SET_TAG(CDR(CDR(x)), install("silent"));
 		DBG(Rprintf("eval(try(.http.request('%s'),silent=TRUE))\n", c->url));
@@ -702,6 +717,7 @@ static void http_input_iteration(args_t *c) {
 								}
 							}
 						}
+						/* lower-case all header names */
 						while (*k && *k != ':') {
 							if (*k >= 'A' && *k <= 'Z')
 								*k |= 0x20;
@@ -719,7 +735,11 @@ static void http_input_iteration(args_t *c) {
 							}
 							if (!strcmp(bol, "content-type")) {
 								char *l = k;
-								while (*l) { if (*l >= 'A' && *l <= 'Z') *l |= 0x20; l++; }
+								/* change the content type to lower case,
+								   however, stop at ; since training content
+								   may be case-sensitive such as multipart-boundary
+								   (see #149) */
+								while (*l && *l != ';') { if (*l >= 'A' && *l <= 'Z') *l |= 0x20; l++; }
 								c->attr |= CONTENT_TYPE;
 								if (c->content_type) free(c->content_type);
 								c->content_type = strdup(k);
@@ -855,6 +875,9 @@ static void http_input_iteration(args_t *c) {
     }
 }
 
+/* from Rserve.c */
+int check_tls_client(int verify, const char *cn);
+
 static void HTTP_connected(void *parg) {
 	args_t *arg = (args_t*) parg;
 
@@ -869,8 +892,16 @@ static void HTTP_connected(void *parg) {
 		return;
 	}
 
-	if ((arg->srv->flags & SRV_TLS) && shared_tls(0))
+	if ((arg->srv->flags & SRV_TLS) && shared_tls(0)) {
+		char cn[256];
 		add_tls(arg, shared_tls(0), 1);
+		if (check_tls_client(verify_peer_tls(arg, cn, 256), cn)) {
+			close_tls(arg);
+			closesocket(arg->s);
+			free_args(arg);
+			return;
+		}
+	}
 
 	while (arg->s != -1)
 		http_input_iteration(arg);
