@@ -5,34 +5,50 @@
 #include "tls.h"
 
 #include "rsdebug.h"
-#include "rserr.h"
 
 #include <sisocks.h>
 #include <string.h>
 #include <stdio.h>
 #include <signal.h>
+#include <stddef.h>
+#include <stdint.h>
+
+#include <Rinternals.h>
+
+#ifdef WIN32
+/* FIXME: this is a terribe, terrible hack around Win32 idiocy.
+   we should use stdint type or some such ... */
+static int getpid() { /* for compatibility */
+	return (int) GetProcessId(GetCurrentProcess());
+}
+#endif
 
 struct args {
 	server_t *srv; /* server that instantiated this connection */
     SOCKET s;
 	SOCKET ss;
+	int msg_id;
 	void *res1; /* used by TLS */
 	struct args *tls_arg; /* if set it is used to wire send/recv calls */
 	/* the following entries are not populated by Rserve but can be used by server implemetations */
 	char *buf, *sbuf;
 	int   ver, bp, bl, sp, sl, flags;
-	long  l1, l2;
+	size_t l1, l2;
 };
 
-static int  WS_recv_data(args_t *arg, void *buf, rlen_t read_len);
-static void WS_send_resp(args_t *arg, int rsp, rlen_t len, const void *buf);
-static int  WS_send_data(args_t *arg, const void *buf, rlen_t len);
+static ssize_t WS_recv_data(args_t *arg, void *buf, size_t read_len);
+static int     WS_send_resp(args_t *arg, int rsp, size_t len, const void *buf);
+static ssize_t WS_send_data(args_t *arg, const void *buf, size_t len);
 
-static int WS_wire_send(args_t *arg, const void *buf, rlen_t len) {
-	return (arg->tls_arg) ? arg->tls_arg->srv->send(arg->tls_arg, buf, len) : send(arg->s, buf, len, 0);
+/* those will eventually be in the API but for now ... */
+ssize_t cio_send(int s, const void *buffer, size_t length, int flags);
+ssize_t cio_recv(int s, void *buffer, size_t length, int flags);
+
+static ssize_t WS_wire_send(args_t *arg, const void *buf, size_t len) {
+	return (arg->tls_arg) ? arg->tls_arg->srv->send(arg->tls_arg, buf, len) : cio_send(arg->s, buf, len, 0);
 }
-static int WS_wire_recv(args_t *arg, void *buf, rlen_t len) {
-	return (arg->tls_arg) ? arg->tls_arg->srv->recv(arg->tls_arg, buf, len) : recv(arg->s, buf, len, 0);
+static ssize_t WS_wire_recv(args_t *arg, void *buf, size_t len) {
+	return (arg->tls_arg) ? arg->tls_arg->srv->recv(arg->tls_arg, buf, len) : cio_recv(arg->s, buf, len, 0);
 }
 static void WS_wire_close(args_t *arg) {
 	if (arg->tls_arg) {
@@ -50,8 +66,8 @@ static void WS_wire_close(args_t *arg) {
 	arg->s = -1;
 }
 
-static int do_mask(char *msg, int len, int koff, char *key) {
-	int i = 0;
+static int do_mask(char *msg, size_t len, int koff, char *key) {
+	size_t i = 0;
 	while (i < len) {
 		msg[i] ^= key[(i + koff) & 3];
 		i++;
@@ -85,15 +101,15 @@ static void free_header(struct header_info *h) {
 	if (h->protocol) free(h->protocol);
 }
 
-static unsigned long count_spaces(const char *c) {
-	unsigned long n = 0;
+static size_t count_spaces(const char *c) {
+	size_t n = 0;
 	while (*c) { if (*c == ' ') n++; c++; }
 	return n;	
 }
 
 static unsigned long count_digits(const char *c) {
-	unsigned long n = 0;
-	while (*c) { if (*c >= '0' && *c <= '9') n = n * 10L + (unsigned long)(*c - '0'); c++; }
+	size_t n = 0;
+	while (*c) { if (*c >= '0' && *c <= '9') n = n * 10L + (size_t)(*c - '0'); c++; }
 	return n;	
 }
 
@@ -103,12 +119,12 @@ void base64encode(const unsigned char *src, int len, char *dst);
 void Rserve_QAP1_connected(args_t *arg);
 void Rserve_text_connected(args_t *arg);
 
+int check_tls_client(int verify, const char *cn);
+
 #define FRAME_BUFFER_SIZE 65536
 
 static void WS_connected(void *parg) {
 	args_t *arg = (args_t*) parg;
-	/* server_t *srv = arg->srv; */
-	SOCKET s = arg->s;
 	int n, bp = 0, empty_lines = 0, request_line = 1;
 
 	struct header_info h;
@@ -118,6 +134,8 @@ static void WS_connected(void *parg) {
 	/* we have to perform a handshake before giving over to QAP 
 	   but we have to fork() first as to not block the server on handshake */
 	if (Rserve_prepare_child(arg) != 0) { /* parent or error */
+		if (arg->s != -1)
+			closesocket(arg->s);
 		free(arg);
 		return;
 	}
@@ -128,10 +146,20 @@ static void WS_connected(void *parg) {
 	   no bad side-effects.
 	*/
 	if (arg->srv->flags & WS_TLS) {
+		char cn[256];
 		args_t *tls_arg = calloc(1, sizeof(args_t));
 		tls_arg->s = arg->s;
 		tls_arg->srv = calloc(1, sizeof(server_t));
 		add_tls(tls_arg, shared_tls(0), 1);
+		if (check_tls_client(verify_peer_tls(tls_arg, cn, 256), cn)) {
+			close_tls(tls_arg);
+			free(tls_arg->srv);
+			free(tls_arg);
+			if (arg->s != -1)
+				closesocket(arg->s);
+			free(arg);
+			return;
+		}
 		arg->tls_arg = tls_arg;
 	} else arg->tls_arg = 0;
 
@@ -141,7 +169,7 @@ static void WS_connected(void *parg) {
 		strcpy(lbuf, "HTTP/1.1 500 Out of memory\r\n\r\n");
 		WS_wire_send(arg, lbuf, strlen(lbuf));
 		WS_wire_close(arg);
-		arg->s = -1;
+		free(arg);
 		return;
 	}
 	buf[LINE_BUF_SIZE - 1] = 0;
@@ -225,6 +253,7 @@ static void WS_connected(void *parg) {
 				arg->s = -1;
 				free_header(&h);
 				free(buf);
+				free(arg);
 				return;
 			}
 			/* otherwise it's fine, we will load more */
@@ -246,6 +275,7 @@ static void WS_connected(void *parg) {
 		arg->s = -1;
 		free(buf);
 		free_header(&h);
+		free(arg);
 		return;
 	}
 #ifdef RSERV_DEBUG
@@ -269,6 +299,7 @@ static void WS_connected(void *parg) {
 				arg->s = -1;
 				free(buf);
 				free_header(&h);
+				free(arg);
 				return;
 			}
 		}
@@ -279,6 +310,7 @@ static void WS_connected(void *parg) {
 			arg->s = -1;
 			free(buf);
 			free_header(&h);
+			free(arg);
 			return;
 		}
 		v[0] = count_digits(h.key1) / count_spaces(h.key1);
@@ -406,20 +438,28 @@ void WS13_upgrade(args_t *arg, const char *key, const char *protocol, const char
 	Rserve_QAP1_connected(arg);
 }
 
-static void WS_send_resp(args_t *arg, int rsp, rlen_t len, const void *buf) {
+static int WS_send_resp(args_t *arg, int rsp, size_t len, const void *buf) {
 	unsigned char *sbuf = (unsigned char*) arg->sbuf;
+	if (len > rlen_max - 128) {
+#ifdef RSERV_DEBUG
+		fprintf(stderr, "WS_send_resp: payload too big (would exceed rlen_max)\n");
+#endif
+		return -1;
+	}
 	if (arg->ver == 0) {
 		/* FIXME: we can't really tunnel QAP1 without some encoding ... */
 	} else {
 		struct phdr ph;
-		int pl = 0;
-		long flen = len + sizeof(ph);
-		memset(&ph, 0, sizeof(ph));
+		size_t pl = 0;
+		size_t flen = len + sizeof(ph);
 		ph.cmd = itop(rsp | ((rsp & CMD_OOB) ? 0 : CMD_RESP));
 		ph.len = itop(len);
 #ifdef __LP64__
 		ph.res = itop(len >> 32);
+#else
+		ph.res = 0;
 #endif
+		ph.msg_id = arg->msg_id;
 
 #ifdef RSERV_DEBUG
 		if (io_log) {
@@ -447,50 +487,65 @@ static void WS_send_resp(args_t *arg, int rsp, rlen_t len, const void *buf) {
 			sbuf[pl++] = 126;
 			sbuf[pl++] = flen >> 8;
 			sbuf[pl++] = flen & 255;
-		} else { /* 64-bit */
+		} else { /* 64-bit (on 32-bit plarforms the high 4 will be 0) */
 			sbuf[pl++] = 127;
-			{ int i = 8; long l = flen; while (i--) { sbuf[pl + i] = l & 255; l >>= 8; } }
+			{
+				int i = 8;
+				size_t l = flen;
+				while (i--) {
+					sbuf[pl + i] = l & 255;
+					l >>= 8;
+				}
+			}
 			pl += 8;
 		}	
 		memcpy(sbuf + pl, &ph, sizeof(ph));
 		pl += sizeof(ph);
 		while (len + pl) {
-			int n, send_here = (len + pl > arg->sl) ? arg->sl : (len + pl);
+			ssize_t n;
+			/* arg->sl is only 64k so there are no size issues */
+			size_t send_here = (len + pl > arg->sl) ? arg->sl : (len + pl);
 			if (send_here > pl)
 				memcpy(sbuf + pl, buf, send_here - pl);
 			n = WS_wire_send(arg, sbuf, send_here);
 #ifdef RSERV_DEBUG
 			if (pl) {
-				fprintf(stderr, "WS_send_resp: sending 4+ frame (ver %02d), n = %d / %d (of total %ld)\n", arg->ver, n, send_here, flen);
+				fprintf(stderr, "WS_send_resp: sending 4+ frame (ver %02d), n = %ld / %ld (of total %ld)\n", arg->ver, (long) n, (unsigned long)send_here, flen);
+#ifdef WS_DEBUG
 				{ int i, m = send_here; if (m > 100) m = 100; for (i = 0; i < m; i++) fprintf(stderr, " %02x", (int) sbuf[i]); fprintf(stderr,"\n"); }
+#endif
 			} else
-				fprintf(stderr, "WS_send_resp: continuation (%d bytes)\n", n);
+				fprintf(stderr, "WS_send_resp: continuation (%ld bytes)\n", (long) n);
 #endif
 			if (n != send_here) {
 #ifdef RSERV_DEBUG
-				fprintf(stderr, "WS_send_resp: write failed (%d expected, got %d)\n", send_here, n);
+				fprintf(stderr, "WS_send_resp: write failed (%ld expected, got %ld)\n",
+						(long) send_here, (long) n);
 #endif
-				return;
+				return -1;
 			}
 			buf = ((char*)buf) + send_here - pl;
 			len -= send_here - pl;
 			pl = 0;
 		}
 	}
+	return 0;
 }
 
 /* we use send_data only to send the ID string so we don't bother supporting frames bigger than the buffer */
-static int  WS_send_data(args_t *arg, const void *buf, rlen_t len) {
+/* FIXME: not true anymore - it is used for pass-through, so we did implement binary fragmentation */
+static ssize_t WS_send_data(args_t *arg, const void *buf, size_t len) {
 	unsigned char *sbuf = (unsigned char*) arg->sbuf;
 	if (arg->ver == 0) {
 		if (len < arg->sl - 2) {
-			int n;
+			ssize_t n;
 			sbuf[0] = 0;
 			memcpy(sbuf + 1, buf, len);
 			sbuf[len + 1] = 0xff;
 			n = WS_wire_send(arg, sbuf, len + 2);
 #ifdef RSERV_DEBUG
-			fprintf(stderr, "WS_send_data: sending 00 frame, n = %d / %d\n", n, (int) len + 2);
+			fprintf(stderr, "WS_send_data: sending 00 frame, n = %ld / %lu\n",
+					(long) n, (unsigned long) len + 2);
 #endif
 			if (n == len + 2) return len;
 			if (n < len + 2 && n >= len) return len - 1;
@@ -502,47 +557,71 @@ static int  WS_send_data(args_t *arg, const void *buf, rlen_t len) {
 			return -1;
 		}
 	} else {
-		if (len < arg->sl - 8 && len < 65536) {
-			int n, pl = 0;
-			sbuf[pl++] =  ((arg->flags & F_OUT_BIN) ? 1 : 0) + ((arg->ver < 4) ? 0x04 : 0x81); /* text, 4+ has inverted FIN bit */
-			if (len < 126) /* short length */
-				sbuf[pl++] = len;
-			else if (len < 65536) { /* 16-bit */
-				sbuf[pl++] = 126;
-				sbuf[pl++] = len >> 8;
-				sbuf[pl++] = len & 255;
+		size_t total = len, pl = 0;
+		sbuf[pl++] =  ((arg->flags & F_OUT_BIN) ? 1 : 0) + ((arg->ver < 4) ? 0x04 : 0x81); /* text, 4+ has inverted FIN bit */
+		if (len < 126) /* short length */
+			sbuf[pl++] = len;
+		else if (len < 65536) { /* 16-bit */
+			sbuf[pl++] = 126;
+			sbuf[pl++] = len >> 8;
+			sbuf[pl++] = len & 255;
+		} else { /* 64-bit */
+			sbuf[pl++] = 127;
+			{
+				int i = 8;
+				size_t l = len;
+				while (i--) {
+					sbuf[pl + i] = l & 255;
+					l >>= 8;
+				}
 			}
-			/* no masking or other stuff */
-			memcpy(sbuf + pl, buf, len);
-			n = WS_wire_send(arg, sbuf, len + pl);
+			pl += 8;
+		}
+        while (len + pl) {
+            ssize_t n;
+			size_t send_here = (len + pl > arg->sl) ? arg->sl : (len + pl);
+            if (send_here > pl)
+                memcpy(sbuf + pl, buf, send_here - pl);
+            n = WS_wire_send(arg, sbuf, send_here);
 #ifdef RSERV_DEBUG
-			fprintf(stderr, "WS_send_data: sending 4+ frame (ver %02d), n = %d / %d\n", arg->ver, n, (int) len + pl);
+            if (pl) {
+                fprintf(stderr, "WS_send_data: sending 4+ frame (ver %02d), n = %ld / %lu (of total %ld)\n", arg->ver, (long) n, (unsigned long) send_here, (long) len);
+#ifdef WS_DEBUG
+                { int i, m = send_here; if (m > 100) m = 100; for (i = 0; i < m; i++) fprintf(stderr, " %02x", (int) sbuf[i]); fprintf(stderr,"\n"); }
 #endif
-			if (n == len + pl) return len;
-			if (n < len + pl && n >= len) return len - 1;
-			return n;
-		} else {
+            } else
+                fprintf(stderr, "WS_send_data: continuation (%ld bytes)\n", (long) n);
+#endif
+            if (n != send_here) {
 #ifdef RSERV_DEBUG
-			fprintf(stderr, "ERROR in WS_send_data: data too large\n");
+                fprintf(stderr, "WS_send_data: write failed (%ld expected, got %ld)\n",
+						(long) send_here, (long) n);
 #endif
-			return -1;
-		}		
+                return -1;
+            }
+            buf = ((char*)buf) + send_here - pl;
+            len -= send_here - pl;
+            pl = 0;
+        }
+		return total;
 	}
 	return 0;
 }
 
-static int  WS_recv_data(args_t *arg, void *buf, rlen_t read_len) {
+static ssize_t WS_recv_data(args_t *arg, void *buf, size_t read_len) {
 #ifdef RSERV_DEBUG
-	fprintf(stderr, "WS_recv_data for %d (bp = %d)\n", (int) read_len, arg->bp);
+	fprintf(stderr, "WS_recv_data for %ld (bp = %d)\n", (long) read_len, arg->bp);
 #endif
 	if (arg->ver == 0) {
 		/* make sure we have at least one (in frame) or two (oof) bytes in the buffer */
 		int min_size = (arg->flags & F_INFRAME) ? 1 : 2;
 		while (arg->bp < min_size) {
-			int n = WS_wire_recv(arg, arg->buf + arg->bp, arg->bl - arg->bp);
+			ssize_t n = WS_wire_recv(arg, arg->buf + arg->bp, arg->bl - arg->bp);
 #ifdef RSERV_DEBUG
-			fprintf(stderr, "WS_recv_data: needs ver 00 frame, reading %d bytes in addition to %d\n", n, arg->bp);
-			{ int i; fprintf(stderr, "Buffer: "); for (i = 0; i < n; i++) fprintf(stderr, " %02x", (int) (unsigned char) arg->buf[i]); fprintf(stderr,"\n"); }
+			fprintf(stderr, "WS_recv_data: needs ver 00 frame, reading %ld bytes in addition to %d\n", (long) n, arg->bp);
+#ifdef WS_DEBUG
+			{ ssize_t i; fprintf(stderr, "Buffer: "); for (i = 0; i < n; i++) fprintf(stderr, " %02x", (int) (unsigned char) arg->buf[i]); fprintf(stderr,"\n"); }
+#endif
 #endif
 			if (n < 1) return n;
 			arg->bp += n;
@@ -564,7 +643,7 @@ static int  WS_recv_data(args_t *arg, void *buf, rlen_t read_len) {
 		/* NOTE: this is actually always true since we guarantee both F_INFRAME and bp > 0 above */
 	    if ((arg->flags & F_INFRAME) && arg->bp > 0) {
 			unsigned char *b = (unsigned char*) arg->buf;
-			int i = 0;
+			ssize_t i = 0;
 #ifdef RSERV_DEBUG
 			fprintf(stderr, "WS_recv_data: have %d bytes of a frame, requested %d, returning what we have\n", arg->bp, (int) read_len);
 #endif
@@ -608,12 +687,20 @@ static int  WS_recv_data(args_t *arg, void *buf, rlen_t read_len) {
 	}
 	/* make sure we have at least one byte in the buffer */
 	if (arg->bp == 0) {
-		int n = WS_wire_recv(arg, arg->buf, arg->bl);
+		/* don't read past the current frame in case we're in a frame ... */
+		/* FIXME: it shouldn't matter but it appears that we don't handle that case correctly */
+		ssize_t n;
+		size_t max_sz = ((arg->flags & F_INFRAME) && arg->l1) ? arg->l1 : arg->bl;
+		if (max_sz > arg->bl)
+			max_sz = arg->bl;
+		n = WS_wire_recv(arg, arg->buf, max_sz);
 		if (n < 1) return n;
 		arg->bp = n;
 #ifdef RSERV_DEBUG
-		fprintf(stderr, "INFO: WS_recv_data: read %d bytes:\n", n);
-		{ int i; for (i = 0; i < n; i++) fprintf(stderr, " %02x", (int) (unsigned char) arg->buf[i]); fprintf(stderr,"\n"); }
+		fprintf(stderr, "INFO: WS_recv_data: read %ld bytes:\n", (long) n);
+#ifdef WS_DEBUG
+		{ ssize_t i; for (i = 0; i < n; i++) fprintf(stderr, " %02x", (int) (unsigned char) arg->buf[i]); fprintf(stderr,"\n"); }
+#endif
 #endif
 	}
 	if (arg->flags & F_INFRAME) { /* in frame with new content */
@@ -631,9 +718,12 @@ static int  WS_recv_data(args_t *arg, void *buf, rlen_t read_len) {
 		return read_len;
 	} else { /* not in frame - interpret a new frame */
 		unsigned char *fr = (unsigned char*) arg->buf;
-		int more = (arg->ver < 4) ? ((fr[0] & 0x80) == 0x80) : ((fr[0] & 0x80) == 0), mask = 0;
-		int need = 0, ct = fr[0] & 127, at_least, payload;
-		long len = 0;
+#ifdef RSERV_DEBUG /* FIXME: we don't use more -- why? */
+		size_t more = (arg->ver < 4) ? ((fr[0] & 0x80) == 0x80) : ((fr[0] & 0x80) == 0);
+#endif
+		int mask = 0, ct = fr[0] & 127;
+		size_t need = 0, at_least, payload;
+		size_t len = 0;
 		/* set the F_IN_BIN flag according to the frame type */
 		if ((arg->ver < 4 && ct == 5) ||
 			(arg->ver >= 4 && ct == 2))
@@ -642,7 +732,7 @@ static int  WS_recv_data(args_t *arg, void *buf, rlen_t read_len) {
 			arg->flags &= ~ F_IN_BIN;
 		SET_F_FT(arg->flags, ct);
 		if (arg->bp == 1) {
-			int n = WS_wire_recv(arg, arg->buf + 1, arg->bl - 1);
+			ssize_t n = WS_wire_recv(arg, arg->buf + 1, arg->bl - 1);
 			if (n < 1) return n;
 			arg->bp = n + 1;
 		}
@@ -650,7 +740,7 @@ static int  WS_recv_data(args_t *arg, void *buf, rlen_t read_len) {
 		len = fr[1] & 127;
 		need = 2 + (mask ? 4 : 0) + ((len < 126) ? 0 : ((len == 126) ? 2 : 8));
 		while (arg->bp < need) {
-			int n = WS_wire_recv(arg, arg->buf + arg->bp, arg->bl - arg->bp);
+			ssize_t n = WS_wire_recv(arg, arg->buf + arg->bp, arg->bl - arg->bp);
 			if (n < 1) return n;
 			arg->bp += n;
 		}
@@ -663,21 +753,31 @@ static int  WS_recv_data(args_t *arg, void *buf, rlen_t read_len) {
 #endif
 				return -1;
 			}
-#define SH(X,Y) (((long)X) << Y)
-			len = SH(fr[4], 48) | SH(fr[5], 40) | SH(fr[5], 32) | SH(fr[6], 24) | SH(fr[7], 16) | SH(fr[8], 8) | (long)fr[9];
+#define SH(X,Y) (((uint64_t)X) << Y)
+			uint64_t raw_len = SH(fr[4], 48) | SH(fr[5], 40) | SH(fr[5], 32) | SH(fr[6], 24) | SH(fr[7], 16) | SH(fr[8], 8) | (uint64_t)fr[9];
+			if (raw_len > rlen_max - 64L) {
+#ifdef RSERV_DEBUG
+				fprintf(stderr, "WS_recv_data: requested frame length is larger than rlen_max for this platform\n");
+#endif
+				return -1;
+			}
+			len = (size_t) raw_len;
 		}
 #ifdef RSERV_DEBUG
-		fprintf(stderr, "INFO: WS_recv_data frame type=%02x, len=%d, more=%d, mask=%d (need=%d)\n", ct, (int) len, more, mask, need);
+		fprintf(stderr, "INFO: WS_recv_data frame type=%02x, len=%lu, more=%d, mask=%d (need=%d)\n", ct, (unsigned long) len, (int) more, (int) mask, (int) need);
 #endif
 		at_least = need + len;
 		if (at_least > arg->bl)
 			at_least = arg->bl;
+
+		/* this is always positive, because need is at most 14 and at_least
+		   is need + len only capped by the buffer size which is >14 */
 		payload = at_least - need;
-		
+
 		while (arg->bp < at_least) {
-			int n = WS_wire_recv(arg, arg->buf + arg->bp, at_least - arg->bp);
+			ssize_t n = WS_wire_recv(arg, arg->buf + arg->bp, at_least - arg->bp);
 #ifdef RSERV_DEBUG
-			fprintf(stderr, "INFO: read extra %d bytes in addition to %d (need %d)\n", n, arg->bp, need);
+			fprintf(stderr, "INFO: read extra %ld bytes in addition to %d (need %ld)\n", (long) n, arg->bp, (long) need);
 #endif
 			if (n < 1) return n;
 			arg->bp += n;
@@ -685,12 +785,17 @@ static int  WS_recv_data(args_t *arg, void *buf, rlen_t read_len) {
 		/* FIXME: more recent protocols require MASK at all times */
 		if (mask) {
 #ifdef RSERV_DEBUG
-			{ int i; for (i = 0; i < payload; i++) fprintf(stderr, " %02x", (int) (unsigned char) arg->buf[need + i]); fprintf(stderr,"\n"); }
+#ifdef WS_DEBUG
+			{ size_t i; for (i = 0; i < payload; i++) fprintf(stderr, " %02x", (int) (unsigned char) arg->buf[need + i]); fprintf(stderr,"\n"); }
+#endif
 #endif
 			SET_F_MASK(arg->flags, do_mask(arg->buf + need, payload, 0, arg->buf + need - 4));
+			/* we (ab)use l2 for the key - its really just char[4] which long can always fit */
 			memcpy(&arg->l2, arg->buf + need - 4, 4);
 #ifdef RSERV_DEBUG
-			{ int i; for (i = 0; i < payload; i++) fprintf(stderr, " %02x", (int) (unsigned char) arg->buf[need + i]); fprintf(stderr,"\n"); }
+#ifdef WS_DEBUG
+			{ size_t i; for (i = 0; i < payload; i++) fprintf(stderr, " %02x", (int) (unsigned char) arg->buf[need + i]); fprintf(stderr,"\n"); }
+#endif
 #endif
 		} else arg->flags &= ~ F_MASK;
 		
@@ -709,10 +814,12 @@ static int  WS_recv_data(args_t *arg, void *buf, rlen_t read_len) {
 		
 		/* left-over */
 #ifdef RSERV_DEBUG
-		fprintf(stderr, "INFO: WS_recv_data frame has %d bytes (of %ld frame), requested %d, returning partial frame\n", payload, len, (int) read_len);
+		fprintf(stderr, "INFO: WS_recv_data frame has %lu bytes (of %lu frame), requested %d, returning partial frame\n",
+				(unsigned long) payload, (unsigned long) len, (int) read_len);
 #endif
 		/* we can only read all we got */
-		if (read_len > payload) read_len = payload;
+		if (read_len > payload)
+			read_len = payload;
 		memcpy(buf, arg->buf + need, read_len);
 		if (arg->bp > need + read_len) /* if there is any data beyond what we will deliver, we need to move it */
 			memmove(arg->buf, arg->buf + need + read_len, arg->bp - need - read_len);
@@ -738,8 +845,6 @@ server_t *create_WS_server(int port, int flags) {
 	return 0;
 }
 
-#include <Rinternals.h>
-
 void serverLoop(void);
 
 typedef void (*sig_fn_t)(int);
